#define PURPLE_PLUGINS

/*
 *
 * Pidgin manual entry area height sizing plugin
 * License: GPLv2, see http://www.fsf.org/ for a full text
 *
 * Copyright (C) 2008-2009, Artemy Kapitula <dalt74@gmail.com>
 *
 * Version 0.7
 *
 */

#include <string.h>
#include "pidgin.h"
#include "gtkprefs.h"

#include "conversation.h"
#include "prefs.h"
#include "signals.h"
#include "version.h"
#include "debug.h"

#include "gtkplugin.h"
#include "gtkutils.h"
#include "gtkimhtml.h"

#ifndef _WIN32
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif

#define N_(A)		A
#define MANUALSIZE_PLUGIN_ID "pidgin-entry-manual-height"
#define PLUGIN_VERSION "0.7"

//
// The size that should be set
//
static int im_new_size = -1;
static int chat_new_size = -1;

static PurplePlugin * my_plugin = NULL;

//
// Conversation consolidated structure.
// Gives to us GtkVPaned, PidginConversation, PidginConversation->lower_hbox and
// their "shown" state in one place. I'm too lazy to figure out their crossdependencies.
//
typedef struct _conventry {
	int entry_shown;
	int ischat;
	GtkPaned * vpaned;
	GtkWidget * lower_hbox;
	PidginConversation * conv;
} ConvEntry;

//
// List of known conversations
// used to implement all-together-resize and
// to implement a clear plugin unload (unregister all installed callbacks)
//
static GList * entries = NULL;

//
// Derecursor. If it's set it means we doing a programm-initiated resize
// after user moved splitter in the initiator->vpaned
//
static ConvEntry * initiator = NULL;

//
// Comparator to use in find_conv_entry()
//
static gint
conv_entry_compare(gconstpointer listdata, gconstpointer needle) {
	ConvEntry * dt = (ConvEntry*)listdata;
	if (dt==needle) return 0;
	if (dt->conv==needle) return 0;
	if (dt->lower_hbox==needle) return 0;
	if (dt->vpaned==needle) return 0;
	if (dt->conv->active_conv==needle) return 0;
	return 1;
}

//
// Find convEntry that refers specified object
// Specified object may be any of
//    ConvEntry itself, ConvEntry->lower_hbox,
//    ConvEntry->vpaned, ConvEntry->conv or ConvEntry->conv->active_conv
//
ConvEntry * find_conv_entry(void * cv) {
	GList * found = g_list_find_custom(entries,cv,conv_entry_compare);
	if (found) return (ConvEntry*)(found->data);
	return NULL;
}

//
// Allocates and initialize ConvEntry
//
ConvEntry * alloc_conv_entry(GtkPaned * vpaned, GtkWidget * lower_hbox, PidginConversation * conv) {
	ConvEntry * ret = (ConvEntry*)malloc(sizeof(ConvEntry));
	ret->entry_shown = 0;
	ret->ischat = (conv->active_conv->type == PURPLE_CONV_TYPE_CHAT);
	ret->vpaned = vpaned;
	ret->lower_hbox = lower_hbox;
	ret->conv = conv;
	return ret;
}

//
// Detects ConvEntry's lower_hbox size
//
int current_lower_size(ConvEntry * c) {
	return c->lower_hbox->allocation.height;
}

//
// Unified "resize" procedure, operates on ConvEntry
//
void set_lower_size(ConvEntry * c, int newsize) {
	int curpos = gtk_paned_get_position(c->vpaned);
	int cursize = current_lower_size(c);
	int newpos = curpos+cursize-newsize;
	if (curpos!=newpos) {
		/*
		printf(
			"Conversation %s: cursize=%d newsize=%d, curpos=%d newpos=%d\n",
			c->conv->active_conv->name, cursize, newsize, curpos, newpos
		);
		*/
		gtk_paned_set_position( c->vpaned,newpos);
	}
}

// Manipulating with new window size
int get_new_size(int chat) {
	if (chat) {
		return chat_new_size;
	} else {
		return im_new_size;
	}
}

void set_new_size(int chat, int s) {
	if (chat) {
		chat_new_size = s;
		purple_prefs_set_int( "/plugins/manualsize/chat_entry_height" , s );
	} else {
		im_new_size = s;
		purple_prefs_set_int( "/plugins/manualsize/im_entry_height" , s );
	}
}

// Iterator for changing size of multiple lower_hbox'es
static void
change_all_sizes_callback(gpointer data, gpointer in_chat) {
	ConvEntry * ce = (ConvEntry*)data;
	if (ce==initiator) return;
	if (
		((ce->ischat==1)&&(in_chat==(gpointer)1)) ||
		((ce->ischat==0)&&(in_chat==(gpointer)0))
	) {
		set_lower_size(ce, get_new_size(ce->ischat));
	}
}

// Change all chat or all im lower_hbox'es height
static void
change_all_sizes(int in_chat) {
	g_list_foreach( entries, change_all_sizes_callback, (gpointer)in_chat );
}

//
// Callback - called when lower_hbox resized.
// Stores entry area height into a preferences.
//
static void lower_hbox_resized( GtkWidget * w, GtkAllocation * a, gpointer conv ) {
	if (initiator) return;
	/* printf("Got \"size-allocate\" signal on lower_hbox\n"); */
	ConvEntry * ce = find_conv_entry(w);
	if (ce->entry_shown==0) {
		/* printf("Startup resize for %s!\n", ce->conv->active_conv->name); */
		set_lower_size(ce,get_new_size(ce->ischat));
	} else {
		if (current_lower_size(ce)!=get_new_size(ce->ischat)) {
			/* printf("User-initiated resize %s!\n", ce->conv->active_conv->name); */
			set_new_size(ce->ischat,current_lower_size(ce));
		}
		initiator = ce;
		if (
			(ce->ischat)&&(purple_prefs_get_bool( "/plugins/manualsize/chat_entry_height_synchronized")) ||
			(ce->ischat==0)&&(purple_prefs_get_bool( "/plugins/manualsize/im_entry_height_synchronized"))
		) {
			change_all_sizes(ce->ischat);
		}
		initiator = NULL;
	}
	ce->entry_shown = 1;
}

//
// Register PidginConversation and connect a handler to the lower_hbox of the conversation
//
static void
connect_resize_handler(GtkPaned * vpaned, GtkWidget * lower_hbox, PidginConversation * conv) {
	entries = g_list_append( entries, alloc_conv_entry(vpaned,lower_hbox,conv) );
	g_signal_connect_after( lower_hbox, "size-allocate", (GCallback)lower_hbox_resized, 0 );
}

//
// Rebuild conversation pane.
// Find a conversation pane ("pane")
// Find a parent for a pane ("top")
// Create GtkVPaned ("vpaned")
// Move "pane" from a "top" to the up of "vpaned"
// Move "lower_hbox" of conversation to the bottom "vpaned"
// Insert "vpaned" into a "top"
// Change "vpaned" divider position
// Set focus where it had before we start
//
static void
rebuild_container(PidginConversation * conv) {

//	GtkWidget * window;		// Conversation window
	GtkWidget * now_focused;	// Currently focused widget
	GtkWidget * pane;		// Conversation pane
	GtkWidget * top;		// Parent of conversation pane
	GtkWidget * vpaned;		// Our GtkVPaned
//	GtkNotebook * notebook;		// Notebook in a conversation window
	gboolean chat;
	int handle_size = 0;
	int parent_area = 0;
	int border_size = 0;
	int new_pos;
	int stored_height;
	GtkPositionType tabpos = -1;
	GValue v;

	pane = gtk_widget_get_parent(GTK_WIDGET(conv->lower_hbox));

	top = gtk_widget_get_parent( pane );

	vpaned= gtk_vpaned_new();

	chat = (conv->active_conv->type == PURPLE_CONV_TYPE_CHAT);
		now_focused = NULL;

	stored_height = (chat)?
		purple_prefs_get_int( "/plugins/manualsize/chat_entry_height" )
		:
		purple_prefs_get_int( "/plugins/manualsize/im_entry_height" );

	if (stored_height < 0) stored_height = 128;

	memset( &v, 0, sizeof(v) );
	g_value_init( &v, G_TYPE_BOOLEAN );
	
	gtk_widget_show( vpaned );

	g_value_set_boolean( &v, TRUE );
	gtk_widget_reparent( pane, vpaned );
	gtk_container_child_set_property( GTK_CONTAINER(vpaned), pane, "resize", &v );

	g_value_set_boolean( &v, FALSE );
	gtk_widget_reparent( conv->lower_hbox, vpaned );
	gtk_container_child_set_property( GTK_CONTAINER(vpaned), conv->lower_hbox, "resize", &v );

	g_value_unset( &v );

	gtk_container_add( GTK_CONTAINER(top), vpaned );

	// Take it back to the widget that has been focused when we start
	if (now_focused) {
		gtk_widget_grab_focus( now_focused );
	} else {
		gtk_widget_grab_focus( conv->entry );
	}

	connect_resize_handler( GTK_PANED(vpaned), conv->lower_hbox, conv );

}

//
// Primary signal handler. Called when conversation created, and rebuilds a conversation pane
//
static void
on_display(void* data) {
	PidginConversation * gtkconv = (PidginConversation*)data;
	rebuild_container( gtkconv );
}

//
// Conversation destroy signal handler.
// Called when conversation destroyed to remove out handlers and to store an input area size
//
static void
on_destroy(void * data) {
	ConvEntry * ce = find_conv_entry(data);
	int chat;

	if (ce) {
		g_signal_handlers_disconnect_by_func( ce->lower_hbox, lower_hbox_resized, 0 );
		entries = g_list_remove( entries, ce );
		if (purple_prefs_get_bool( "/plugins/manualsize/dont_save_entry_height_on_exit")==FALSE)
			set_new_size(ce->ischat,current_lower_size(ce));
		free(ce);
	} else {
		printf("It shouldn't be happend! It's a bug!\n");
	}
}

static gboolean
plugin_load(PurplePlugin *plugin)
{
	void * gtk_conv_handle = pidgin_conversations_get_handle();
	void * conv_handle = purple_conversations_get_handle();

//	myself = plugin;
	purple_prefs_add_none( "/plugins" );
	purple_prefs_add_none( "/plugins/manualsize" );
	purple_prefs_add_int( "/plugins/manualsize/chat_entry_height", 128 );
	purple_prefs_add_int( "/plugins/manualsize/im_entry_height", 128 );
	purple_prefs_add_bool( "/plugins/manualsize/im_entry_height_synchronized", FALSE );
	purple_prefs_add_bool( "/plugins/manualsize/chat_entry_height_synchronized", FALSE );
	purple_prefs_add_bool( "/plugins/manualsize/dont_save_entry_height_on_exit", FALSE );

	set_new_size(1,purple_prefs_get_int( "/plugins/manualsize/chat_entry_height" ));
	set_new_size(0,purple_prefs_get_int( "/plugins/manualsize/im_entry_height" ));

	my_plugin = plugin;

	purple_signal_connect(gtk_conv_handle, "conversation-displayed", plugin, PURPLE_CALLBACK(on_display), NULL);
	purple_signal_connect(conv_handle, "deleting-conversation", plugin, PURPLE_CALLBACK(on_destroy), NULL);

	return TRUE;
}

static void
cleanup_convs_callback(gpointer data, gpointer user_data) {
	ConvEntry * ce = (ConvEntry*)data;
	g_signal_handlers_disconnect_by_func( ce->lower_hbox, lower_hbox_resized, 0);
}

static gboolean
plugin_unload(PurplePlugin *plugin)
{
	g_list_foreach(entries, cleanup_convs_callback, NULL );
	g_list_free(entries);
	return TRUE;
}

static GtkWidget *
get_config_frame(PurplePlugin *plugin)
{
	GtkWidget *ret;
	ret = gtk_vbox_new(FALSE, PIDGIN_HIG_CAT_SPACE);
	pidgin_prefs_checkbox(N_("Change entry area height in all IM conversations at once"),
	                        "/plugins/manualsize/im_entry_height_synchronized", ret);
	pidgin_prefs_checkbox(N_("Change entry area height in all chat conversations at once"),
	                        "/plugins/manualsize/chat_entry_height_synchronized", ret);
	pidgin_prefs_checkbox(N_("Don't save entry height on exit, save after resize only"),
	                        "/plugins/manualsize/dont_save_entry_height_on_exit", ret);
	gtk_widget_show_all(ret);
	return ret;
}

static PidginPluginUiInfo ui_info =
{
	get_config_frame,
	0, /* page_num (Reserved) */

	/* padding */
	NULL,
	NULL,
	NULL,
	NULL
};

static PurplePluginInfo info =
{
	PURPLE_PLUGIN_MAGIC,
	PURPLE_MAJOR_VERSION,
	PURPLE_MINOR_VERSION,
	PURPLE_PLUGIN_STANDARD,                           /**< type           */
	PIDGIN_PLUGIN_TYPE,                               /**< ui_requirement */
	0,                                                /**< flags          */
	NULL,                                             /**< dependencies   */
	PURPLE_PRIORITY_DEFAULT,                          /**< priority       */

	MANUALSIZE_PLUGIN_ID,                             /**< id             */
	N_("Entry area manual sizing"),                   /**< name           */
	PLUGIN_VERSION,                                   /**< version        */
	                                                  
	N_("Allows you to change entry area height"),     /**  summary        */
	N_("Allows you to change entry area height"),     /**  description    */
	                                                  
	"Artemy Kapitula <dalt74@gmail.com>",             /**< author         */
	"http://myfotomx.com/builds_en.html",             /**< homepage       */

	plugin_load,                                      /**< load           */
	plugin_unload,                                    /**< unload         */
	NULL,                                             /**< destroy        */

	&ui_info,                                         /**< ui_info        */
	NULL,                                             /**< extra_info     */
	NULL,
	NULL,

	/* padding */
	NULL,
	NULL,
	NULL,
	NULL
};

static void
init_plugin(PurplePlugin *plugin)
{
    return;
}

PURPLE_INIT_PLUGIN(manualsize, init_plugin, info)

